// src/farcaster/frames.tsx
import {
  FarcasterNetwork,
  FrameActionBody,
  Message,
  NobleEd25519Signer,
  makeFrameAction
} from "@farcaster/core";
import { hexToBytes } from "viem";
var signFrameAction = async ({
  buttonIndex,
  frameContext,
  frameButton,
  signer,
  target,
  inputText,
  state,
  transactionId,
  url
}) => {
  if (!signer || signer.status === "pending_approval") {
    throw new Error("Missing signer fid to sign message");
  }
  const { message, trustedBytes } = await createFrameActionMessageWithSignerKey(
    signer.privateKey,
    {
      fid: signer.fid,
      buttonIndex,
      castId: {
        fid: frameContext.castId.fid,
        hash: hexToBytes(frameContext.castId.hash)
      },
      state: state !== void 0 ? Buffer.from(state) : void 0,
      url: Buffer.from(url),
      // it seems the message in hubs actually requires a value here.
      inputText: inputText !== void 0 ? Buffer.from(inputText) : void 0,
      address: frameContext.address !== void 0 ? hexToBytes(frameContext.address) : void 0,
      transactionId: transactionId !== void 0 ? hexToBytes(transactionId) : void 0
    }
  );
  if (!message) {
    throw new Error("hub error");
  }
  const searchParams = new URLSearchParams({
    postType: transactionId ? "post" : frameButton.action,
    postUrl: target ?? ""
  });
  return {
    searchParams,
    body: {
      untrustedData: {
        fid: signer.fid,
        url,
        messageHash: `0x${Buffer.from(message.hash).toString("hex")}`,
        timestamp: message.data.timestamp,
        network: 1,
        buttonIndex: Number(message.data.frameActionBody.buttonIndex),
        castId: {
          fid: frameContext.castId.fid,
          hash: frameContext.castId.hash
        },
        inputText,
        address: frameContext.address,
        transactionId,
        state
      },
      trustedData: {
        messageBytes: trustedBytes
      }
    }
  };
};
async function createFrameActionMessageWithSignerKey(signerKey, {
  fid,
  url,
  buttonIndex,
  castId,
  inputText,
  state,
  address,
  transactionId
}) {
  const signer = new NobleEd25519Signer(Buffer.from(signerKey.slice(2), "hex"));
  const messageDataOptions = {
    fid,
    network: FarcasterNetwork.MAINNET
  };
  const message = await makeFrameAction(
    FrameActionBody.create({
      url,
      buttonIndex,
      castId,
      state,
      inputText: inputText !== void 0 ? Buffer.from(inputText) : void 0,
      address,
      transactionId
    }),
    messageDataOptions,
    signer
  );
  if (message.isErr()) {
    return { message: null, trustedBytes: null };
  }
  const trustedBytes = Buffer.from(
    Message.encode(message._unsafeUnwrap()).finish()
  ).toString("hex");
  return { message: message.unwrapOr(null), trustedBytes };
}

export {
  signFrameAction,
  createFrameActionMessageWithSignerKey
};
