import {
  isParseResult
} from "./chunk-23M226IE.js";

// src/use-fetch-frame.ts
function computeDurationInSeconds(start, end) {
  return Number(((end.getTime() - start.getTime()) / 1e3).toFixed(2));
}
function defaultErrorHandler(error) {
  console.error(error);
}
function useFetchFrame({
  stackDispatch,
  specification,
  frameActionProxy,
  frameGetProxy,
  extraButtonRequestPayload,
  signFrameAction,
  onTransaction,
  homeframeUrl,
  onError = defaultErrorHandler,
  fetchFn,
  onRedirect
}) {
  async function handleFailedResponse({
    response,
    startTime,
    endTime,
    frameStackPendingItem
  }) {
    if (response.ok) {
      throw new TypeError(
        "handleFailedResponse called with a successful response"
      );
    }
    const responseBody = await getResponseBody(response);
    if (response.status >= 400 && response.status < 500) {
      if (frameStackPendingItem.request.method === "POST" && responseBody && typeof responseBody === "object" && "message" in responseBody && typeof responseBody.message === "string") {
        const stackItem = {
          ...frameStackPendingItem,
          responseStatus: response.status,
          response: response.clone(),
          speed: computeDurationInSeconds(startTime, endTime),
          status: "message",
          type: "error",
          message: responseBody.message,
          responseBody
        };
        stackDispatch({
          action: "DONE",
          pendingItem: frameStackPendingItem,
          item: stackItem
        });
        return;
      }
      console.error(
        `frames.js: The server returned an error but it does not contain message property. Status code: ${response.status}`,
        responseBody
      );
    }
    const requestError = new Error(
      `The server returned an error but it does not contain message property. Status code: ${response.status}`
    );
    pushRequestErrorToStack({
      frameStackPendingItem,
      requestError,
      responseStatus: response.status,
      startTime,
      endTime,
      responseBody,
      response
    });
  }
  function pushRequestErrorToStack({
    frameStackPendingItem,
    requestError,
    responseStatus,
    startTime,
    endTime,
    responseBody,
    response
  }) {
    stackDispatch({
      action: "REQUEST_ERROR",
      pendingItem: frameStackPendingItem,
      item: {
        request: frameStackPendingItem.request,
        requestDetails: frameStackPendingItem.requestDetails,
        timestamp: frameStackPendingItem.timestamp,
        url: frameStackPendingItem.url,
        response: response?.clone() ?? null,
        responseStatus,
        requestError,
        speed: computeDurationInSeconds(startTime, endTime),
        status: "requestError",
        responseBody
      }
    });
    onError(requestError);
  }
  async function fetchGETRequest(request, shouldClear) {
    const startTime = /* @__PURE__ */ new Date();
    if (shouldClear) {
      stackDispatch({ action: "CLEAR" });
    }
    const frameStackPendingItem = {
      method: "GET",
      timestamp: startTime,
      request,
      status: "pending",
      requestDetails: {},
      url: request.url
    };
    stackDispatch({ action: "LOAD", item: frameStackPendingItem });
    const searchParams = new URLSearchParams({
      url: request.url,
      specification
    });
    const proxiedUrl = `${frameGetProxy}?${searchParams.toString()}`;
    const response = await tryCall(fetchFn(proxiedUrl, { method: "GET" }));
    const endTime = /* @__PURE__ */ new Date();
    if (response instanceof Response) {
      if (!response.ok) {
        await handleFailedResponse({
          response,
          startTime,
          endTime,
          frameStackPendingItem
        });
        return;
      }
      const loadedFrame = await response.clone().json();
      stackDispatch({
        action: "DONE",
        pendingItem: frameStackPendingItem,
        item: {
          ...frameStackPendingItem,
          status: "done",
          frameResult: loadedFrame,
          speed: computeDurationInSeconds(startTime, endTime),
          response: response.clone(),
          responseStatus: response.status,
          responseBody: loadedFrame
        }
      });
      return;
    }
    pushRequestErrorToStack({
      frameStackPendingItem,
      endTime,
      startTime,
      responseBody: "none",
      responseStatus: 500,
      requestError: response,
      response: null
    });
  }
  async function fetchPOSTRequest(request, options, shouldClear) {
    const { address: _, ...requiredFrameContext } = request.signerStateActionContext.frameContext;
    const startTime = options?.startTime || /* @__PURE__ */ new Date();
    const frameStackPendingItem = {
      method: "POST",
      request,
      requestDetails: {
        // this will be filled afterwards
      },
      timestamp: startTime,
      status: "pending",
      url: request.frameButton.target || request.sourceFrame.postUrl || homeframeUrl || ""
      // this will be also filled after action is signed
    };
    if (shouldClear) {
      stackDispatch({ action: "CLEAR" });
    }
    if (!options) {
      stackDispatch({ action: "LOAD", item: frameStackPendingItem });
    }
    const signedDataOrError = await tryCall(
      signFrameAction(request.isDangerousSkipSigning, {
        ...request.signerStateActionContext,
        frameContext: requiredFrameContext
      })
    );
    if (signedDataOrError instanceof Error) {
      pushRequestErrorToStack({
        endTime: /* @__PURE__ */ new Date(),
        frameStackPendingItem,
        responseBody: "none",
        responseStatus: 500,
        startTime,
        requestError: signedDataOrError,
        response: null
      });
      return;
    }
    const { body, searchParams } = signedDataOrError;
    stackDispatch({
      action: "ADD_REQUEST_DETAILS",
      requestDetails: {
        body,
        searchParams
      },
      url: searchParams.get("postUrl") ?? "",
      pendingItem: frameStackPendingItem
    });
    searchParams.set("specification", specification);
    const proxiedUrl = `${frameActionProxy}?${searchParams.toString()}`;
    const response = await tryCall(
      fetchFn(proxiedUrl, {
        method: "POST",
        headers: {
          "Content-Type": "application/json"
        },
        body: JSON.stringify({
          ...extraButtonRequestPayload,
          ...body
        })
      })
    );
    const endTime = /* @__PURE__ */ new Date();
    async function handleRedirect(res, pendingItem) {
      try {
        let location = res.headers.get("location");
        if (!location) {
          const responseData = await res.clone().json();
          if (responseData && typeof responseData === "object" && "location" in responseData && typeof responseData.location === "string") {
            location = responseData.location;
          }
        }
        if (!location) {
          throw new Error(
            `Response data does not contain 'location' key and no 'location' header is found.`
          );
        }
        const locationUrl = new URL(location);
        onRedirect(locationUrl);
        stackDispatch({
          action: "DONE_REDIRECT",
          pendingItem,
          item: {
            ...pendingItem,
            location,
            response: res.clone(),
            responseBody: await res.clone().text(),
            responseStatus: res.status,
            status: "doneRedirect",
            speed: computeDurationInSeconds(startTime, endTime)
          }
        });
      } catch (e) {
        stackDispatch({
          action: "REQUEST_ERROR",
          pendingItem: frameStackPendingItem,
          item: {
            ...frameStackPendingItem,
            status: "requestError",
            requestError: e instanceof Error ? e : new Error(
              "Response body must be a json with 'location' property or response 'Location' header must contain fully qualified URL."
            ),
            response: res.clone(),
            responseStatus: res.status,
            responseBody: await res.text(),
            speed: computeDurationInSeconds(startTime, endTime)
          }
        });
      }
    }
    if (response instanceof Response) {
      if (response.status === 302) {
        await handleRedirect(response, frameStackPendingItem);
        return;
      }
      if (!response.ok) {
        await handleFailedResponse({
          response,
          startTime,
          endTime,
          frameStackPendingItem
        });
        return;
      }
      const responseData = await response.clone().json();
      if ("message" in responseData) {
        stackDispatch({
          action: "DONE",
          pendingItem: frameStackPendingItem,
          item: {
            ...frameStackPendingItem,
            status: "message",
            message: responseData.message,
            type: "info",
            response: response.clone(),
            responseStatus: response.status,
            speed: computeDurationInSeconds(startTime, endTime),
            responseBody: responseData
          }
        });
        return;
      }
      if ("frameUrl" in responseData) {
        await fetchPOSTRequest(
          {
            // actions don't have source frame, fake it
            sourceFrame: {
              image: "",
              version: "vNext"
            },
            frameButton: {
              action: "post",
              label: "action",
              target: responseData.frameUrl
            },
            isDangerousSkipSigning: request.isDangerousSkipSigning,
            method: "POST",
            signerStateActionContext: {
              ...request.signerStateActionContext,
              buttonIndex: 1,
              frameButton: {
                action: "post",
                label: "action",
                target: responseData.frameUrl
              },
              target: responseData.frameUrl
            }
          },
          {
            pendingFrameStackItem: frameStackPendingItem,
            startTime
          }
        );
        return;
      }
      if (!isParseResult(responseData)) {
        const error = new Error(`The server returned an unexpected response.`);
        console.error(error);
        pushRequestErrorToStack({
          endTime,
          frameStackPendingItem,
          responseBody: responseData,
          responseStatus: 500,
          startTime,
          requestError: error,
          response
        });
        return;
      }
      stackDispatch({
        action: "DONE",
        pendingItem: options?.pendingFrameStackItem ?? frameStackPendingItem,
        item: {
          ...options?.pendingFrameStackItem ?? frameStackPendingItem,
          // update url to the one used to send the request, this handles transaction flow where pending item is created for transaction data request
          url: frameStackPendingItem.url,
          frameResult: responseData,
          status: "done",
          speed: computeDurationInSeconds(startTime, endTime),
          response: response.clone(),
          responseStatus: response.status,
          responseBody: responseData
        }
      });
      return;
    }
    pushRequestErrorToStack({
      endTime,
      frameStackPendingItem,
      responseBody: "none",
      responseStatus: 500,
      startTime,
      requestError: response,
      response: new Response(response.message, {
        status: 500,
        headers: { "Content-Type": "text/plain" }
      })
    });
  }
  async function fetchTransactionRequest(request, shouldClear) {
    const button = request.frameButton;
    if (button.action !== "tx") {
      throw new Error("Invalid frame button action, tx expected");
    }
    if (shouldClear) {
      stackDispatch({ action: "CLEAR" });
    }
    const startTime = /* @__PURE__ */ new Date();
    const frameStackPendingItem = {
      method: "POST",
      timestamp: startTime,
      status: "pending",
      request,
      // will be filled afterwards
      requestDetails: {},
      url: request.frameButton.target || request.sourceFrame.postUrl || homeframeUrl || ""
    };
    stackDispatch({ action: "LOAD", item: frameStackPendingItem });
    const signedTransactionDataActionOrError = await tryCall(
      signFrameAction(
        // for transaction data we always use signer, so skip signing is false here
        false,
        request.signerStateActionContext
      )
    );
    if (signedTransactionDataActionOrError instanceof Error) {
      pushRequestErrorToStack({
        endTime: /* @__PURE__ */ new Date(),
        frameStackPendingItem,
        responseBody: "none",
        responseStatus: 500,
        startTime,
        requestError: signedTransactionDataActionOrError,
        response: null
      });
      return;
    }
    stackDispatch({
      action: "ADD_REQUEST_DETAILS",
      requestDetails: {
        body: signedTransactionDataActionOrError.body,
        searchParams: signedTransactionDataActionOrError.searchParams
      },
      url: signedTransactionDataActionOrError.searchParams.get("postUrl") ?? "",
      pendingItem: frameStackPendingItem
    });
    signedTransactionDataActionOrError.searchParams.set(
      "specification",
      specification
    );
    const transactionDataProxiedUrl = `${frameActionProxy}?${signedTransactionDataActionOrError.searchParams.toString()}`;
    const transactionDataResponse = await tryCall(
      fetchFn(transactionDataProxiedUrl, {
        method: "POST",
        headers: {
          "Content-Type": "application/json"
        },
        body: JSON.stringify({
          ...extraButtonRequestPayload,
          ...signedTransactionDataActionOrError.body
        })
      })
    );
    const endTime = /* @__PURE__ */ new Date();
    if (!(transactionDataResponse instanceof Response)) {
      pushRequestErrorToStack({
        endTime,
        frameStackPendingItem,
        responseBody: "none",
        responseStatus: 500,
        startTime,
        requestError: transactionDataResponse,
        response: null
      });
      return;
    }
    if (!transactionDataResponse.ok) {
      await handleFailedResponse({
        response: transactionDataResponse,
        startTime,
        endTime,
        frameStackPendingItem
      });
      return;
    }
    const transactionData = await transactionDataResponse.json();
    const transactionIdOrError = await tryCall(
      onTransaction({
        frame: request.sourceFrame,
        frameButton: request.frameButton,
        transactionData
      }).then((transactionId) => {
        if (!transactionId) {
          return new Error("onTransaction did not return transaction id");
        }
        return transactionId;
      })
    );
    if (transactionIdOrError instanceof Error) {
      pushRequestErrorToStack({
        endTime,
        frameStackPendingItem,
        responseBody: "none",
        responseStatus: 500,
        startTime,
        requestError: transactionIdOrError,
        response: null
      });
      return;
    }
    await fetchPOSTRequest(
      {
        ...request,
        signerStateActionContext: {
          ...request.signerStateActionContext,
          // include transactionId in payload
          transactionId: transactionIdOrError,
          // override target so the the request is sent to proper endpoint
          target: button.post_url || request.sourceFrame.postUrl || button.target
        }
      },
      {
        pendingFrameStackItem: frameStackPendingItem,
        startTime
      }
    );
  }
  return (request, shouldClear = false) => {
    if (request.method === "GET") {
      return fetchGETRequest(request, shouldClear);
    }
    if (request.frameButton.action === "tx") {
      return fetchTransactionRequest(request, shouldClear);
    }
    return fetchPOSTRequest(request, void 0, shouldClear);
  };
}
async function tryCall(promise) {
  return promise.catch((e) => {
    if (e instanceof Error) {
      return e;
    }
    console.error(e);
    return new TypeError("Unexpected error, check the console for details");
  });
}
function getResponseBody(response) {
  if (response.headers.get("content-type")?.includes("/json")) {
    return response.clone().json();
  }
  return response.clone().text();
}

export {
  useFetchFrame
};
