import { FrameButton, SupportedParsingSpecification, Frame, FrameButtonLink, FrameButtonPost, FrameButtonTx, getFrame, TransactionTargetResponse } from 'frames.js';
import { Dispatch } from 'react';
import { ParseResult } from 'frames.js/frame-parsers';
import { CastId, FrameActionMessage } from '@farcaster/core';

type FarcasterSignerState = SignerStateInstance<FarcasterSigner | null>;
type FarcasterSignerPendingApproval = {
    status: "pending_approval";
    deadline: number;
    privateKey: string;
    publicKey: string;
    requestFid: number;
    requestSigner: string;
    signature: string;
    signerApprovalUrl: string;
    token: string;
};
type FarcasterSignerImpersonating = {
    status: "impersonating";
    fid: number;
    privateKey: string;
    publicKey: string;
};
type FarcasterSignerApproved = {
    status: "approved";
    fid: number;
    privateKey: string;
    publicKey: string;
};
type FarcasterSigner = FarcasterSignerPendingApproval | FarcasterSignerImpersonating | FarcasterSignerApproved;
declare const mockFarcasterSigner: FarcasterSigner;

type FarcasterFrameContext = {
    /** Connected address of user, only sent with transaction data request */
    address?: `0x${string}`;
    castId: {
        hash: `0x${string}`;
        fid: number;
    };
};
/** Creates a frame action for use with `useFrame` and a proxy */
declare const signFrameAction: ({ buttonIndex, frameContext, frameButton, signer, target, inputText, state, transactionId, url, }: {
    target?: string;
    frameButton: FrameButton;
    buttonIndex: number;
    url: string;
    signer: FarcasterSignerState["signer"];
    inputText?: string;
    state?: string;
    transactionId?: `0x${string}`;
    frameContext: FarcasterFrameContext;
}) => Promise<{
    body: FrameActionBodyPayload;
    searchParams: URLSearchParams;
}>;
declare function createFrameActionMessageWithSignerKey(signerKey: string, { fid, url, buttonIndex, castId, inputText, state, address, transactionId, }: {
    fid: number;
    url: Uint8Array;
    buttonIndex: number;
    inputText: Uint8Array | undefined;
    castId: CastId;
    state: Uint8Array | undefined;
    address: Uint8Array | undefined;
    transactionId: Uint8Array | undefined;
}): Promise<{
    message: null;
    trustedBytes: null;
} | {
    message: FrameActionMessage | null;
    trustedBytes: string;
}>;

type OnTransactionFunc = (t: OnTransactionArgs) => Promise<`0x${string}` | null>;
type UseFetchFrameOptions = {
    stackDispatch: React.Dispatch<FrameReducerActions>;
    specification: SupportedParsingSpecification;
    /**
     * URL or path to the frame proxy handling GET requests.
     */
    frameGetProxy: string;
    /**
     * URL or path to the frame proxy handling POST requests.
     */
    frameActionProxy: string;
    /**
     * Extra payload to be sent with the POST request.
     */
    extraButtonRequestPayload?: Record<string, unknown>;
    signFrameAction: (isDangerousSkipSigning: boolean, actionContext: SignerStateActionContext<any, any>) => ReturnType<SignerStateInstance["signFrameAction"]>;
    onTransaction: OnTransactionFunc;
    homeframeUrl: string | undefined | null;
    /**
     * This function can be used to customize how error is reported to the user.
     */
    onError?: (error: Error) => void;
    /**
     * Custom fetch compatible function used to make requests.
     *
     * @see https://developer.mozilla.org/en-US/docs/Web/API/Fetch_API
     */
    fetchFn: typeof fetch;
    /**
     * This function is called when the frame returns a redirect in response to post_redirect button click.
     */
    onRedirect: (location: URL) => void;
};
type UseFrameOptions<SignerStorageType = object, FrameActionBodyType extends FrameActionBodyPayload = FrameActionBodyPayload, FrameContextType extends FrameContext = FarcasterFrameContext> = {
    /** skip frame signing, for frames that don't verify signatures */
    dangerousSkipSigning?: boolean;
    /** the route used to POST frame actions. The post_url will be added as a the `url` query parameter */
    frameActionProxy: string;
    /** the route used to GET the initial frame via proxy */
    frameGetProxy: string;
    /** an signer state object used to determine what actions are possible */
    signerState: SignerStateInstance<SignerStorageType, FrameActionBodyType, FrameContextType>;
    /** the url of the homeframe, if null / undefined won't load a frame */
    homeframeUrl: string | null | undefined;
    /** the initial frame. if not specified will fetch it from the homeframeUrl prop */
    frame?: Frame | ParseResult;
    /** connected wallet address of the user */
    connectedAddress: `0x${string}` | undefined;
    /** a function to handle mint buttons */
    onMint?: (t: OnMintArgs) => void;
    /** a function to handle transaction buttons, returns the transaction hash or null */
    onTransaction?: OnTransactionFunc;
    /** the context of this frame, used for generating Frame Action payloads */
    frameContext: FrameContextType;
    /**
     * Extra data appended to the frame action payload
     */
    extraButtonRequestPayload?: Record<string, unknown>;
    /**
     * Which specification to use for parsing the frame action payload
     *
     * @defaultValue 'farcaster'
     */
    specification?: SupportedParsingSpecification;
    /**
     * This function can be used to customize how error is reported to the user.
     */
    onError?: (error: Error) => void;
    /**
     * This function can be used to customize how the link button click is handled.
     */
    onLinkButtonClick?: (button: FrameButtonLink) => void;
} & Partial<Pick<UseFetchFrameOptions, "fetchFn" | "onRedirect">>;
type SignerStateActionContext<SignerStorageType = object, FrameContextType extends FrameContext = FarcasterFrameContext> = {
    target?: string;
    frameButton: FrameButton;
    buttonIndex: number;
    url: string;
    inputText?: string;
    signer: SignerStorageType | null;
    state?: string;
    transactionId?: `0x${string}`;
    address?: `0x${string}`;
    /** Transacting address is not included in non-transaction frame actions */
    frameContext: FrameContextType | Omit<FrameContextType, "address">;
};
interface SignerStateInstance<SignerStorageType = object, FrameActionBodyType extends FrameActionBodyPayload = FrameActionBodyPayload, FrameContextType extends FrameContext = FarcasterFrameContext> {
    signer?: SignerStorageType | null;
    hasSigner: boolean;
    signFrameAction: (actionContext: SignerStateActionContext<SignerStorageType, FrameContextType>) => Promise<{
        body: FrameActionBodyType;
        searchParams: URLSearchParams;
    }>;
    /** isLoading frame */
    isLoading?: null | FrameStackPending;
    /** is loading the signer */
    isLoadingSigner?: boolean;
    /** A function called when a frame button is clicked without a signer */
    onSignerlessFramePress: () => void;
    logout?: () => void;
}
type FrameGETRequest = {
    method: "GET";
    url: string;
};
type FramePOSTRequest = {
    method: "POST";
    frameButton: FrameButtonPost | FrameButtonTx;
    signerStateActionContext: SignerStateActionContext<any, any>;
    isDangerousSkipSigning: boolean;
    /**
     * The frame that was the source of the button press
     */
    sourceFrame: Frame;
};
type FrameRequest = FrameGETRequest | FramePOSTRequest;
type FrameStackBase = {
    timestamp: Date;
    /** speed in seconds */
    speed: number;
    responseStatus: number;
    responseBody: unknown;
    requestDetails: {
        body?: object;
        searchParams?: URLSearchParams;
    };
    url: string;
};
type FrameStackPostPending = {
    method: "POST";
    timestamp: Date;
    status: "pending";
    request: FramePOSTRequest;
    requestDetails: {
        body?: object;
        searchParams?: URLSearchParams;
    };
    url: string;
};
type FrameStackGetPending = {
    method: "GET";
    timestamp: Date;
    status: "pending";
    request: FrameGETRequest;
    requestDetails: {
        body?: object;
        searchParams?: URLSearchParams;
    };
    url: string;
};
type FrameStackPending = FrameStackGetPending | FrameStackPostPending;
type GetFrameResult = ReturnType<typeof getFrame>;
type FrameStackDone = FrameStackBase & {
    request: FrameRequest;
    response: Response;
    frameResult: GetFrameResult;
    status: "done";
};
type FrameStackDoneRedirect = FrameStackBase & {
    request: FramePOSTRequest;
    response: Response;
    location: string;
    status: "doneRedirect";
};
type FrameStackRequestError = FrameStackBase & {
    request: FrameRequest;
    response: Response | null;
    status: "requestError";
    requestError: Error;
};
type FrameStackMessage = FrameStackBase & {
    request: FramePOSTRequest;
    response: Response;
    status: "message";
    message: string;
    type: "info" | "error";
};
type FramesStackItem = FrameStackPending | FrameStackDone | FrameStackDoneRedirect | FrameStackRequestError | FrameStackMessage;
type FramesStack = FramesStackItem[];
type FrameReducerActions = {
    action: "LOAD";
    item: FrameStackPending;
} | {
    action: "ADD_REQUEST_DETAILS";
    pendingItem: FrameStackPending;
    requestDetails: {
        body?: object;
        searchParams?: URLSearchParams;
    };
    url: string;
} | {
    action: "REQUEST_ERROR";
    pendingItem: FrameStackPending;
    item: FrameStackRequestError;
} | {
    action: "DONE_REDIRECT";
    pendingItem: FrameStackPending;
    item: FrameStackDoneRedirect;
} | {
    action: "DONE";
    pendingItem: FrameStackPending;
    item: FramesStackItem;
} | {
    action: "CLEAR";
} | {
    action: "RESET_INITIAL_FRAME";
    resultOrFrame: ParseResult | Frame;
    homeframeUrl: string | null | undefined;
};
type FrameState = {
    fetchFrame: (request: FrameRequest, 
    /**
     * If true, the frame stack will be cleared before the new frame is loaded
     *
     * @defaultValue false
     */
    shouldClear?: boolean) => Promise<void>;
    clearFrameStack: () => void;
    dispatchFrameStack: Dispatch<FrameReducerActions>;
    /** The frame at the top of the stack (at index 0) */
    currentFrameStackItem: FramesStackItem | undefined;
    /** A stack of frames with additional context, with the most recent frame at index 0 */
    framesStack: FramesStack;
    inputText: string;
    setInputText: (s: string) => void;
    onButtonPress: (frame: Frame, frameButton: FrameButton, index: number, fetchFrameOverride?: (request: FrameRequest) => Promise<void>) => void | Promise<void>;
    homeframeUrl: string | null | undefined;
};
type OnMintArgs = {
    target: string;
    frameButton: FrameButton;
    frame: Frame;
};
type OnTransactionArgs = {
    transactionData: TransactionTargetResponse;
    frameButton: FrameButton;
    frame: Frame;
};
declare const themeParams: readonly ["bg", "buttonColor", "buttonBg", "buttonBorderColor", "buttonRadius", "buttonHoverBg"];
type FrameTheme = Partial<Record<(typeof themeParams)[number], string>>;
type FrameActionBodyPayload = Record<string, unknown>;
type FrameContext = Record<string, unknown>;

export { type OnTransactionArgs as A, themeParams as B, type FrameTheme as C, type FrameActionBodyPayload as D, type FrameContext as E, type FarcasterFrameContext as F, type GetFrameResult as G, type OnTransactionFunc as O, type SignerStateActionContext as S, type UseFetchFrameOptions as U, type FarcasterSignerState as a, type FarcasterSignerPendingApproval as b, createFrameActionMessageWithSignerKey as c, type FarcasterSignerImpersonating as d, type FarcasterSignerApproved as e, type FarcasterSigner as f, type UseFrameOptions as g, type SignerStateInstance as h, type FrameGETRequest as i, type FramePOSTRequest as j, type FrameRequest as k, type FrameStackBase as l, mockFarcasterSigner as m, type FrameStackPostPending as n, type FrameStackGetPending as o, type FrameStackPending as p, type FrameStackDone as q, type FrameStackDoneRedirect as r, signFrameAction as s, type FrameStackRequestError as t, type FrameStackMessage as u, type FramesStackItem as v, type FramesStack as w, type FrameReducerActions as x, type FrameState as y, type OnMintArgs as z };
